﻿module Noise
open System
open OpenTK


type vec3 = Vector3d
type ArrayList<'T> = System.Collections.Generic.List<'T>

type PerlinNoise() =
    let B = 256
    let BM,DBN,N = B-1,2*B + 2,B*B
    //let B,BM,N,NM,NP,DBN=256,255,4096,4095,12,(2*256+2) 
    //permutation 排列表 实现一种伪随机的效果
    let p = Array.create DBN 0
    //gradient 梯度表 索引在排列表中
    let pg = Array.create DBN vec3.Zero
    let curve t =
        // t * t * (3. - 2. * t) ) 
        t * t * t * (t * (t * 6. - 15.) + 10.)         
    let lerp t a b = a + t*(b-a)
    //得到点vec在对应维度上(x,y,z)上的二边的索引，以及二边的贡献点，二阶平滑插值
    let pos vec =
        //去掉负值 1/B*floor(P)+ONEHALF; 
        let t = vec + float N + 0.5 // Math.Floor(float vec)/(float B) + 1.0/(float (2*B))
        //let t = Math.Floor(float vec)/(float B) + 1.0/(float (2*B))
        //点d的左边,前面,下面索引  (指的是排列表中的位置索引)
        let x0 = int (Math.Floor(t)) &&& BM
        //点d的右边,右面,上面索引
        let x1 = x0 + 1 &&& BM
        //点d在当前维度上,用于对应梯度中左边(x轴),前面(z轴),下面(y轴)各点贡献值
        let u0 = t - Math.Floor(t)
         //其中u0+(-u1)=1.0,-(1.0-u0) 右边,后边,上面各点贡献值 (
        let u1 = u0 - 1.0   
        //二阶平滑位置
        let cf = curve u0
        x0,x1,u0,u1,cf   
    let init() =
        let rand = Random(2)
        for i = 0 to BM do
            p.[i] <- i
            pg.[i] <- vec3(rand.NextDouble()*2.0 - 1.0,rand.NextDouble()*2.0 - 1.0,rand.NextDouble()*2.0 - 1.0)
            //pg.[i]  <- vec2( float (rand.Next()), float (rand.Next()))
            pg.[B+i] <- pg.[i]
        for i = 0 to BM do
            let j = (rand.Next() >>> 5) % BM
            let t = p.[i]
            p.[i] <- p.[j]
            p.[j] <- t
            p.[i + B] <- p.[i]           
    do
        init()        
    member this.Noise1 (x:float) =
        let x0,x1,u0,u1,cx = pos x
        //得到二点的梯度值.
        let ug0 = u0 * pg.[p.[x0]].X
        //-(1.0-f0) 负表示
        let ug1 = u1 * pg.[p.[x1]].X
        //u v 平滑位置 noise1位置
        lerp cx ug0 ug1    
    member this.Noise2 (x,y) =
        let x0,x1,u0,u1,cx = pos x
        let y0,y1,v0,v1,cy = pos y
        //2维中权重也需要混合 下标DBN，p.[x0] + y0 最大为2B
        let gx0y0 = pg.[p.[p.[x0] + y0]] 
        let gx1y0 = pg.[p.[p.[x1] + y0]] 
        let gx0y1 = pg.[p.[p.[x0] + y1]] 
        let gx1y1 = pg.[p.[p.[x1] + y1]] 
        let mixuv (v3:vec3) u v = v3.X * u + v3.Y * v
        //yo轴上x混合值
        let u0v0 = mixuv gx0y0 u0 v0 //gx0y0.X * u0 + gx0y0.Y * v0
        let u1v0 = mixuv gx1y0 u1 v0//gx1y0.X * u1 + gx1y0.Y * v0
        let ccx0 = lerp cx u0v0 u1v0
        //y1轴上x混合值
        let u0v1 = mixuv gx0y1 u0 v1//gx0y1.X * u0 + gx0y1.Y * v1
        let u1v1 = mixuv gx1y1 u1 v1//gx1y1.X * u1 + gx1y1.Y * v1
        let ccx1 = lerp cx u0v1 u1v1
        //混合二y轴
        let result = lerp cy ccx0 ccx1
        // -1 --- 1映射成 0 --- 1
        result //* 0.5 + 0.5
    member this.Noise3 (x,y,z) =
        let x0,x1,u0,u1,cx = pos x
        let y0,y1,v0,v1,cy = pos y
        let z0,z1,r0,r1,cz = pos z
        //权重与当前点(一维左右二个,二维上下左右四个,三维八个)上各分量的点积
        let mixuvr (v3:vec3) u v r = v3.X * u + v3.Y * v + v3.Z * r
        //3维中权重也需要混合 zo面 同二维
        let gx0y0z0 = pg.[p.[p.[x0] + y0] + z0]
        let gx1y0z0 = pg.[p.[p.[x1] + y0] + z0]
        let gx0y1z0 = pg.[p.[p.[x0] + y1] + z0]
        let gx1y1z0 = pg.[p.[p.[x1] + y1] + z0]   

        let u0v0r0 = mixuvr gx0y0z0 u0 v0 r0
        let u1v0r0 = mixuvr gx1y0z0 u1 v0 r0
        let ccx0z0 = lerp cx u0v0r0 u1v0r0

        let u0v1r0 = mixuvr gx0y1z0 u0 v1 r0
        let u1v1r0 = mixuvr gx1y1z0 u1 v1 r0
        let ccx1z0 = lerp cx u0v1r0 u1v1r0    
        
        let ccyz0 = lerp cy ccx0z0 ccx1z0
        //z1面 同二维
        let gx0y0z1 = pg.[p.[p.[x0] + y0] + z1]
        let gx1y0z1 = pg.[p.[p.[x1] + y0] + z1]
        let gx0y1z1 = pg.[p.[p.[x0] + y1] + z1]
        let gx1y1z1 = pg.[p.[p.[x1] + y1] + z1]       
        
        let u0v0r1 = mixuvr gx0y0z1 u0 v0 r1
        let u1v0r1 = mixuvr gx1y0z1 u1 v0 r1
        let ccx0z1 = lerp cx u0v0r1 u1v0r1

        let u0v1r1 = mixuvr gx0y1z1 u0 v1 r1
        let u1v1r1 = mixuvr gx1y1z1 u1 v1 r1
        let ccx1z1 = lerp cx u0v1r1 u1v1r1 
          
        let ccyz1 = lerp cy ccx0z1 ccx1z1
        //混合z轴
        lerp cz ccyz0 ccyz1     
    //octave 倍频,amplitude振幅,调整高低,frequency频率,越小间隔越平滑
    member this.PerlinNoise2(x,y,octave,amplitude,frequency) =
        let mutable total = 0.0
        for i in [|0 .. octave - 1|] do
            let freq  = Math.Pow(frequency,float i)
            let amp = Math.Pow(amplitude,float i)
            total <- total + this.Noise2(x*freq,y*freq)*amp
        total
    member this.PerlinNoise3(x,y,z,octave,amplitude,frequency) =
        let mutable total = 0.0
        for i in [|0 .. octave - 1|] do
            let freq  = Math.Pow(frequency,float i)
            let amp = Math.Pow(amplitude,float i)
            total <- total + this.Noise3(x*freq,y*freq,z*freq)*amp
        total    
    //生成一张图片RGBA源,用于测试
    member this.CreateTexture() =        
        init()
        let size = B * B
        //生成二维纹理GRBA,纹理长与度分别为B。
        let data = Array2D.init size 4 (fun yx i ->
           let y = yx / 256 //每列索引
           let x = yx % 256 //每行数据索引
           let offest = yx * 4
           let vp = p.[p.[x] + y]
           let vg = pg.[vp] * 64.0 + vec3(64.0,64.0,64.0)//由-1到1,映射成0-128
           if i = 0 then byte (vg.X)
           else if i = 1 then byte (vg.Y)
           else if i = 1 then byte (vg.Z)
           else byte (vg.X)
           )
        data
    //生成一张图片RGBA源,用于测试
    member this.Test() =        
        init()
        let arrayrgb = ArrayList<byte>()
        let mutable maxvalue = 0.0
        for i in [| 0 .. 255|] do
            for j in [| 0 .. 255|] do
               // let r = this.Noise2 (float i + float i/256.,float j + float j/256.) 
                let r = this.Noise2 (float i,float j ) 
                //let r = this.Noise2 (float i/256.,float j/256. ) 
                let color = byte (r*128.) + 128uy
                arrayrgb.AddRange([| color; color; color;color |])         
        arrayrgb.ToArray()
       
//如果我们想要坐标(i,j,k)的gradient g(i,j,k),而P里预存储了256个gradient, 那么:
//	g(i, j, k) = P[ ( i + P[ (j + P[k]) mod 256 ] ) mod 256 ]


//    let p = [| 
//        151;160;137;91;90;15;  
//        131;13;201;95;96;53;194;233;7;225;140;36;103;30;69;142;8;99;37;240;21;10;23;  
//        190; 6;148;247;120;234;75;0;26;197;62;94;252;219;203;117;35;11;32;57;177;33;  
//        88;237;149;56;87;174;20;125;136;171;168; 68;175;74;165;71;134;139;48;27;166;  
//        77;146;158;231;83;111;229;122;60;211;133;230;220;105;92;41;55;46;245;40;244;  
//        102;143;54; 65;25;63;161; 1;216;80;73;209;76;132;187;208; 89;18;169;200;196;  
//        135;130;116;188;159;86;164;100;109;198;173;186; 3;64;52;217;226;250;124;123;  
//        5;202;38;147;118;126;255;82;85;212;207;206;59;227;47;16;58;17;182;189;28;42;  
//        223;183;170;213;119;248;152; 2;44;154;163; 70;221;153;101;155;167; 43;172;9;  
//        129;22;39;253; 19;98;108;110;79;113;224;232;178;185; 112;104;218;246;97;228;  
//        251;34;242;193;238;210;144;12;191;179;162;241; 81;51;145;235;249;14;239;107;  
//        49;192;214; 31;181;199;106;157;184; 84;204;176;115;121;50;45;127; 4;150;254;  
//        138;236;205;93;222;114;67;29;24;72;243;141;128;195;78;66;215;61;156;180; 
//        151;160;137;91;90;15;  
//        131;13;201;95;96;53;194;233;7;225;140;36;103;30;69;142;8;99;37;240;21;10;23;  
//        190; 6;148;247;120;234;75;0;26;197;62;94;252;219;203;117;35;11;32;57;177;33;  
//        88;237;149;56;87;174;20;125;136;171;168; 68;175;74;165;71;134;139;48;27;166;  
//        77;146;158;231;83;111;229;122;60;211;133;230;220;105;92;41;55;46;245;40;244;  
//        102;143;54; 65;25;63;161; 1;216;80;73;209;76;132;187;208; 89;18;169;200;196;  
//        135;130;116;188;159;86;164;100;109;198;173;186; 3;64;52;217;226;250;124;123;  
//        5;202;38;147;118;126;255;82;85;212;207;206;59;227;47;16;58;17;182;189;28;42;  
//        223;183;170;213;119;248;152; 2;44;154;163; 70;221;153;101;155;167; 43;172;9;  
//        129;22;39;253; 19;98;108;110;79;113;224;232;178;185; 112;104;218;246;97;228;  
//        251;34;242;193;238;210;144;12;191;179;162;241; 81;51;145;235;249;14;239;107;  
//        49;192;214; 31;181;199;106;157;184; 84;204;176;115;121;50;45;127; 4;150;254;  
//        138;236;205;93;222;114;67;29;24;72;243;141;128;195;78;66;215;61;156;180;   
//        |]