﻿module NoiseForm

open System
open System.Collections.Generic
open ObjReadForm
open Md2ReadForm
open Md5ReadForm
open OpenTK
open OpenTK.Graphics
open OpenTK.Graphics.OpenGL
open Common
open System.Drawing
open System.Drawing.Imaging
open Noise
open CgNet.GL

type NoiseWindow() =
    inherit GameWindow(800,600)
    let speed = 180.
    let mutable offest = 0.01f
    let mutable offestd = float offest
    let caram = Camera()
    let mutable oldMouseLocation = Vector2.Zero
    let pn = PerlinNoise()
    let plane = Plane(200,200,0.25f,0.25f)
    let modelPath = Environment.CurrentDirectory;
    let cgContext = new CgContext()
    let mutable texture = 0
    override v.OnLoad e =
        base.OnLoad e
        let version = GL.GetString(StringName.Version)
        let major = int version.[0]
        let minor =int version.[2]
        if major <= 1 && minor < 5 then 
            v.Exit()
        GL.ClearColor(System.Drawing.Color.MidnightBlue)
        GL.Enable(EnableCap.DepthTest)
        GL.Enable(EnableCap.Texture2D)                
        caram.Transelt(0.5,10.5,0.0)
        cgContext.CreateVectexProgram(modelPath + @"\Data\Shaders\v4.cg","main")
        cgContext.CreateFragmentProgram(modelPath + @"\Data\Shaders\v4.cg","fragmentShader")
        GL.GenTextures(1,&texture)
        GL.BindTexture(TextureTarget.Texture2D,texture)
        let data = pn.CreateTexture()
        GL.TexImage2D(TextureTarget.Texture2D,0,PixelInternalFormat.Rgba32f, 256,256, 0,
            OpenTK.Graphics.OpenGL.PixelFormat.Rgba, PixelType.UnsignedByte, data)        
        GL.TexParameter(TextureTarget.Texture2D,TextureParameterName.TextureMinFilter, (int)TextureMinFilter.Linear)
        GL.TexParameter(TextureTarget.Texture2D, TextureParameterName.TextureMagFilter, (int)TextureMagFilter.Linear)
        cgContext.VertexParameter("permTexture").SetTexture(texture)
    override v.OnResize e =
        base.OnResize e
        GL.Viewport(0,0,v.Width,v.Height)
        let aspect = float32 v.Width / float32 v.Height 
        let mutable projection = Matrix4.CreatePerspectiveFieldOfView(MathHelper.PiOver4,aspect,0.01f,264.f)
        GL.MatrixMode(MatrixMode.Projection)
        GL.LoadMatrix(&projection)
    override v.OnUpdateFrame e =
        base.OnUpdateFrame e
        if v.Keyboard.[OpenTK.Input.Key.Escape] then v.Exit()   
        if v.Keyboard.[OpenTK.Input.Key.E] then caram.Transelt(0.,0.,1.* offestd)
        if v.Keyboard.[OpenTK.Input.Key.D] then caram.Transelt(0.,0.,-1.* offestd)
        if v.Keyboard.[OpenTK.Input.Key.S] then caram.Transelt(1.* offestd,0.,0.0)
        if v.Keyboard.[OpenTK.Input.Key.F] then caram.Transelt(-1.* offestd,0.,0.)
        if v.Keyboard.[OpenTK.Input.Key.Space] then 
            if v.Keyboard.[OpenTK.Input.Key.LShift] then caram.Transelt(0.,-1.* offestd,0.)
            else caram.Transelt(0.,1.* offestd,0.)
        if v.Keyboard.[OpenTK.Input.Key.V] then caram.IsThree <- not caram.IsThree
        if v.Keyboard.[OpenTK.Input.Key.Up] then offest <- offest + offest/0.5f
        if v.Keyboard.[OpenTK.Input.Key.Down] then offest <- offest - offest/0.5f
        if v.Mouse.[OpenTK.Input.MouseButton.Right] then
            if oldMouseLocation = Vector2.Zero then
                oldMouseLocation <- Vector2(float32 v.Mouse.X,float32 v.Mouse.Y)
            else
                let nx = (float32 v.Mouse.X - oldMouseLocation.X) * offest 
                let ny = (float32 v.Mouse.Y - oldMouseLocation.Y) * offest 
                caram.Rotate(float nx,float -ny)
        oldMouseLocation <- Vector2(float32 v.Mouse.X,float32 v.Mouse.Y) 
    override v.OnUnload e =
        cgContext.Unload() 
    override v.OnRenderFrame e =
        base.OnRenderFrame e
        GL.Clear(ClearBufferMask.ColorBufferBit ||| ClearBufferMask.DepthBufferBit)
        let mutable lookat = Matrix4.LookAt(caram.Eye,caram.Target,Vector3.UnitY)
        GL.MatrixMode(MatrixMode.Modelview)
        GL.LoadMatrix(&lookat)

        cgContext.EnableProfile()
        let mv = lookat
        mv.Transpose()
        let mutable p = Matrix4.Identity
        GL.GetFloat(GetPName.ProjectionMatrix,&p)
        p.Transpose()
        let mvp = Matrix4.Mult(p,mv)
        cgContext.VertexParameter("mvp").SetMatrix(MatrixToArray1 mvp)
        cgContext.VertexParameter("permTexture").EnableTexture()
        let date = float (DateTime.Now.Millisecond /200) 
        cgContext.VertexParameter("time").Set(date)
        plane.Draw()
        cgContext.VertexParameter("permTexture").DisableTexture()
        cgContext.DisableProfile()

        v.SwapBuffers()
